package com.example.julian.endlessroll.main.game;

import android.view.MotionEvent;
import android.view.View;

import com.example.julian.endlessroll.data.Vector;
import com.example.julian.endlessroll.entities.DestroyEffect;
import com.example.julian.endlessroll.entities.Energy;
import com.example.julian.endlessroll.entities.Player;
import com.example.julian.endlessroll.entities.Star;
import com.example.julian.endlessroll.entities.collision.CollisionManager;
import com.example.julian.endlessroll.entities.particles.ParticleSystem;
import com.example.julian.endlessroll.entities.textures.TexturePack;
import com.example.julian.endlessroll.entities.tools.ToolType;
import com.example.julian.endlessroll.levels.Level;
import com.example.julian.endlessroll.levels.LevelManager;
import com.example.julian.endlessroll.levels.LevelPack;
import com.example.julian.endlessroll.levels.levelup.LevelUpBounties;
import com.example.julian.endlessroll.main.DataStorageHandler;
import com.example.julian.endlessroll.main.GameHandler;
import com.example.julian.endlessroll.main.GameLog;
import com.example.julian.endlessroll.main.screens.Screen;
import com.example.julian.endlessroll.rendering.Rendering;
import com.example.julian.endlessroll.sounds.SoundManager;
import com.example.julian.endlessroll.views.MessageType;
import com.example.julian.endlessroll.views.ToolButton;
import com.example.julian.endlessroll.views.ToolButtonBar;
import com.example.julian.endlessroll.views.TopBarData;
import com.example.julian.endlessroll.views.ViewManager;

import java.util.ArrayList;
import java.util.List;

/**
 * Created by Julian on 26.11.2015.
 */
public class Game extends Rendering<GameScene> {

    private User user;
    private DataStorageHandler dataStorageHandler;
    private LevelManager levelManager;
    private GameHandler handler;
    private ViewManager viewManager;
    private SoundManager sounds;
    private LevelPack levelPack;
    private ParticleSystem particleSystem;
    private LevelUpBounties levelUpBounties;

    private ToolType currentTool;
    private Player player;
    private Physics physics;
    private CollisionManager collisionManager;
    private Timer timer;
    private GameState gameState = GameState.PAUSED;

    private Level level;
    private List<Integer> collectedStars = new ArrayList<>();
    private boolean energyCollected;

    public Game(GameHandler handler, TopBarData topBarData) throws Exception {
        super(topBarData.getGameActivity());
        this.handler = handler;
        this.user = topBarData.getUser();
        this.levelManager = topBarData.getLevelManager();
        this.sounds = topBarData.getSoundManager();
        levelUpBounties = new LevelUpBounties(0);
        physics = new Physics();
        collisionManager = new CollisionManager(this);
        particleSystem = new ParticleSystem(getContext());
        this.dataStorageHandler = topBarData.getDataStorageHandler();
        viewManager = new ViewManager(this, handler, topBarData);
    }

    @Override
    public GameScene init(TexturePack texturePack, Timer timer, boolean isFirstTime) {
        GameLog.d("initGame");
        this.timer = timer;
        try {
            if (isFirstTime) {
                scene = new GameScene(texturePack, particleSystem);
                if (level != null)
                    startGame(levelPack, level);
            } else {
                scene.setTexturePack(texturePack);
            }
            particleSystem.loadTextures();
        } catch (Exception e) {
            onException(e);
        }
        return scene;
    }

    public void startGame(LevelPack levelPack, Level level) {
        GameLog.d("Start game");
        try {
            this.level = level;
            this.levelPack = levelPack;
            if (scene != null) {
                collectedStars.clear();
                particleSystem.deleteAllSources();
                scene.loadLevel(level, levelPack.getTheme());
                player = scene.getPlayer();
                if (viewManager.isScreenSizeSet())
                    viewManager.resetViews();
                currentTool = viewManager.toolButtonBar.getActiveButton().getToolType();
                viewManager.startCountdown();
            }
        } catch (Exception e) {
            onException(e);
        }
    }

    public void countdownFinished() {
        gameState = GameState.RUNNING;
    }

    @Override
    public void setScreenSize(int width, int height) {
        Vector screenSize = new Vector(width, height);
        scene.setScreenSize(screenSize);
        viewManager.setScreenSize(screenSize);
    }

    @Override
    public void update() {
//        GameLog.d("update GAME");
        particleSystem.update(timer);
        float playerProgress = 0;
        float playerSpeed = 0;
        float playerY = 0;
        if (player != null) {
            playerProgress = player.getProgress();
            playerSpeed = player.getSpeed();
            playerY = player.getPosition().y;
        }
        viewManager.update(gameState == GameState.RUNNING, timer, playerProgress, playerSpeed, playerY);
        switch (gameState) {
            case RUNNING:
                if (player.getPosition().y < -2f) {
                    onGameOver(false);
                    return;
                }
                if (player.getPosition().x >= scene.getGoalX()) {
                    onGoalReached();
                    return;
                }
                physics.applyGravity(scene, timer);
                scene.update(timer);
                collisionManager.update(physics, scene);
                break;
        }
    }

    @Override
    public boolean onTouch(View v, MotionEvent event) {
        if (gameState == GameState.RUNNING) {
            ToolButtonBar bar = viewManager.toolButtonBar;
            ToolButton button = bar.getByToolType(currentTool);
            if (button != null && button.finishedLoading()) {
                if (event.getAction() == MotionEvent.ACTION_DOWN) {
                    button.setProgress(0);
                    addTool(event.getX(), event.getY());
                }
            }
            return true;
        }
        return false;
    }

    public void resetViews() {
        viewManager.resetViews();
    }

    public void continueGame() {
        viewManager.hideShortMenu();
        viewManager.startCountdown();
    }

    public void startNextLevel() {
        level = levelPack.getNextLevel(level);
        startGame(levelPack, level);
    }

    public void restartLevel() {
        startGame(levelPack, level);
    }

    public void toLevelsScreen() {
        handler.toScreen(Screen.ScreenType.LEVELS);
    }

    public void setCurrentTool(ToolType toolType) {
        currentTool = toolType;
    }

    public void pause() {
        viewManager.showShortMenu();
        if (gameState == GameState.RUNNING) {
            gameState = GameState.PAUSED;
        } else {
            viewManager.stopCountdown();
        }
    }

    private void addTool(float x, float y) {
        try {
            sounds.playSound(currentTool.getPlacingSound());
            scene.addTool(currentTool, x, y);
        } catch (Exception e) {
            onException(e);
        }
    }

    public void onGameOver(boolean playerExplode) {
        if (playerExplode) {
            scene.remove(player);
            DestroyEffect.EXPLOSION.createEffect(particleSystem, player.getPosition(), new Vector(player.getWidth(), player.getHeight())).start();
        }
        gameState = GameState.GAME_OVER;
        viewManager.showMessage(false, MessageType.GAME_OVER);
    }

    private void onGoalReached() {
        gameState = GameState.LEVEL_FINISHED;
        if (!level.isFinished())
            user.gainLvFinishedEp();
        level.setFinished(true);

        for (int i = 0; i <= 2; i++) {
            if (collectedStars.contains(i)) {
                level.setStarCollected(i, true);
                user.onStarCollected();
            }
        }

        level.setEnergyCollected(true);
        user.onEnergyCollected();

        viewManager.showMessage(levelPack.isLastLevel(level), MessageType.WIN);

        dataStorageHandler.writeLevelProgress(level);
        Level nextLevel = levelPack.getNextLevel(level);
        if (nextLevel != null) {
            nextLevel.setLocked(false);
            dataStorageHandler.writeLevelProgress(nextLevel);
        } else {
            LevelPack nextLevelPack = levelManager.getNextLevelPack(levelPack);
            if (nextLevelPack != null) {
                nextLevelPack.setLocked(false);
                dataStorageHandler.writeLevelPackLocked(nextLevelPack);
            }
        }
        dataStorageHandler.writeUserData(user);
    }

    public void onStarCollision(Star star) {
        scene.onStarCollision(star);
        collectedStars.add(star.getIndex());
    }

    public void onEnergyCollision(Energy energy){
        scene.onEnergyCollision(energy);
        energyCollected = true;
    }

    @Override
    public void onException(Exception e) {
        handler.onException(e);
    }

    public GameState getGameState() {
        return gameState;
    }

}
