package com.example.julian.endlessroll.main.game;

import com.example.julian.endlessroll.data.Vector;
import com.example.julian.endlessroll.entities.DestroyEffect;
import com.example.julian.endlessroll.entities.Energy;
import com.example.julian.endlessroll.entities.Entity;
import com.example.julian.endlessroll.entities.Goal;
import com.example.julian.endlessroll.entities.Obstacle;
import com.example.julian.endlessroll.entities.Star;
import com.example.julian.endlessroll.entities.collision.CollisionDetector;
import com.example.julian.endlessroll.entities.particles.ParticleSystem;
import com.example.julian.endlessroll.entities.textures.TexturePack;
import com.example.julian.endlessroll.entities.tools.Bomb;
import com.example.julian.endlessroll.entities.tools.Tool;
import com.example.julian.endlessroll.entities.tools.ToolType;
import com.example.julian.endlessroll.levels.Level;
import com.example.julian.endlessroll.levels.ObstacleData;
import com.example.julian.endlessroll.levels.PositionData;
import com.example.julian.endlessroll.levels.worlds.Theme;
import com.example.julian.endlessroll.main.GameLog;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Created by Julian on 27.11.2015.
 */
public class GameScene extends Scene {

    private Theme currentTheme;
    private CollisionDetector collisionDetector;
    private Goal goal;

    private float goalX;
    private List<Obstacle> obstacles = Collections.synchronizedList(new ArrayList<Obstacle>());
    private List<Star> stars = Collections.synchronizedList(new ArrayList<Star>());
    private List<Tool> tools = Collections.synchronizedList(new ArrayList<Tool>());
    private Energy energy;

    public GameScene(TexturePack texturePack, ParticleSystem particleSystem) throws Exception {
        super(texturePack, particleSystem);
        collisionDetector = new CollisionDetector();

        goal = new Goal(textures.goal);
    }

    public void loadLevel(Level level, Theme theme) throws Exception {
        this.currentTheme = theme;
        reset();
        background.changeTexture(theme.getBackgroundTexture());
        terrain.loadData(theme, level.getTerrainEdge(), level.getTerrainTiles());
        ceiling.loadData(theme, level.getCeilingEdge(), level.getCeilingTiles());
        super.add(goal);
        player.init(terrain.getEdge(), level.getStartSpeed(), level.getEndSpeed());
        super.add(player);
        tryToAddEnergy(level);
        for (ObstacleData data : level.getObstacles())
            addObstacle(data);
        for (int i = 0; i < level.getStars().size(); i++) {
            if (level.isStarCollected(i))
                continue;
            addStar(i, level.getStars().get(i));
        }
        goalX = level.getGoalX();
        goal.setGoalX(goalX);
        GameLog.d("Level " + level.getId() + " successfully loaded");
    }

    private void tryToAddEnergy(Level level){
        PositionData data = level.getEnergyData();
        if(data != null && !level.isEnergyCollected()) {
            energy = new Energy(textures.energy, data);
            super.add(energy);
        }
    }

    private void reset() {
        super.clear();
        obstacles.clear();
        tools.clear();
        energy = null;
        cameraX = 0;
        background.resetPosition();
    }

    public void onStarCollision(Star collisionStar) {
        synchronized (stars) {
            Iterator<Star> iter = stars.iterator();
            while (iter.hasNext()) {
                Star star = iter.next();
                if (star.equals(collisionStar)) {
                    star.destroy(DestroyEffect.STAR_EXPLOSION);
                    iter.remove();
                }
            }
        }
    }

    public void onEnergyCollision(Energy energy){
        energy.destroy(DestroyEffect.STAR_EXPLOSION);
    }

    public void addObstacle(ObstacleData data) {
        Obstacle obstacle = new Obstacle(currentTheme, data, terrain.getEdge());
        super.add(obstacle);
        obstacles.add(obstacle);
    }

    public void addStar(int index, PositionData data) {
        Star star = new Star(index, textures.star, data);
        super.add(star);
        stars.add(star);
    }

    public void addTool(ToolType type, float screenX, float screenY) throws Exception {
        Vector position = calcWorldFromScreenCoords(screenX, screenY);
        Tool tool = type.newInstance(position, particleSystem);
        if (tool == null)
            throw new Exception("Current ToolType(" + type + ") returns null at method newInstance()");
        super.add(tool);
        tools.add(tool);
    }


    @Override
    public void update(Timer timer) {
        super.update(timer);
        player.setSpeedByProgress(player.getProgress() / goalX);
        synchronized (tools) {
            for (Tool tool : tools) {
                if (tool instanceof Bomb) {
                    Bomb bomb = (Bomb) tool;
                    if (bomb.isExploding())
                        bomb.explode(obstacles, collisionDetector);
                }
            }
        }

    }

    @Override
    protected void removeEntityFromAllLists(Entity entity) {
        GameLog.i("REMOVE FROM ALL LISTS");
        if (!entity.isDestroyed())
            entity.destroy(null);
        if (entity instanceof Tool)
            tools.remove(entity);
        if (entity instanceof Obstacle)
            obstacles.remove(entity);
    }

    public synchronized List<Tool> getTools() {
        return tools;
    }

    public synchronized List<Obstacle> getObstacles() {
        return obstacles;
    }

    public synchronized List<Star> getStars() {
        return stars;
    }

    public Energy getEnergy() {
        return energy;
    }

    public boolean hasEnergy(){
        return energy != null;
    }

    public float getGoalX() {
        return goalX;
    }
}
