package com.example.julian.endlessroll.main.game;

import android.support.annotation.Nullable;

import com.example.julian.endlessroll.entities.Entity;
import com.example.julian.endlessroll.entities.Obstacle;
import com.example.julian.endlessroll.entities.Star;
import com.example.julian.endlessroll.entities.collision.CollisionDetector;
import com.example.julian.endlessroll.entities.collision.collisionData.EntityCollisionData;
import com.example.julian.endlessroll.entities.collision.collisionData.ObstacleCollisionData;
import com.example.julian.endlessroll.entities.collision.collisionData.PlayerCollisionData;
import com.example.julian.endlessroll.entities.collision.collisionData.ToolCollisionData;
import com.example.julian.endlessroll.entities.collision.geometry.Circle;
import com.example.julian.endlessroll.entities.tileLists.Ceiling;
import com.example.julian.endlessroll.entities.tileLists.Terrain;
import com.example.julian.endlessroll.entities.tileLists.Tile;
import com.example.julian.endlessroll.entities.tools.Bomb;
import com.example.julian.endlessroll.entities.tools.Tool;
import com.example.julian.endlessroll.main.GameLog;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created by Julian on 27.11.2015.
 */
public class Physics {

    public final float GRAVITY_FORCE = .0000025f;
    private CollisionDetector detector;

    public Physics() {
        detector = new CollisionDetector();
    }

    public void applyGravity(GameScene scene, Timer timer) {
        float gravity = GRAVITY_FORCE * timer.getFrameTime();

        scene.getPlayer().getMovement().y -= gravity;
        synchronized (scene.getTools()) {
            for (Tool tool : scene.getTools()) {
                if (tool.isFloating())
                    continue;
                if (tool instanceof Bomb) {
                    for (Obstacle obstacle : scene.getObstacles()) {
                        if (detector.quadQuadCollision(obstacle, tool)) {
                            tool.setFloating(true);
                            tool.getMovement().y = 0;
                            break;
                        }
                    }
                    if (tool.isFloating())
                        continue;
                }
                float terrainEdge = getTerrainEdge(tool, scene.getTerrain());
                Obstacle toolIsOver = getObstacleToolIsOver(tool, scene.getObstacles());

                float orientingHeight = terrainEdge;
                if (toolIsOver != null) {
                    if (toolIsOver.getBottomEdge() < tool.getTopEdge())
                        orientingHeight = Math.max(toolIsOver.getTopEdge(), terrainEdge);
                    else {
                        orientingHeight = terrainEdge;
                    }
                }

                if (tool.getBottomEdge() > orientingHeight) {
                    tool.getMovement().y -= gravity * 2;
                } else {
                    tool.getMovement().y = 0;
                    tool.setToTerrain(orientingHeight);
                }
            }
        }
    }

    @Nullable
    private Obstacle getObstacleToolIsOver(Entity tool, List<Obstacle> obstacles) {
        Map<Float, Obstacle> isOver = new HashMap<>();
        synchronized (obstacles) {
            for (Obstacle obstacle : obstacles) {
                boolean toolVertexInObstacleEdge = (tool.getLeftEdge() >= obstacle.getLeftEdge() && tool.getLeftEdge() <= obstacle.getRightEdge()) || (tool.getRightEdge() <= obstacle.getRightEdge() && tool.getRightEdge() >= obstacle.getLeftEdge());
                boolean obstacleVertexInToolEdge = (obstacle.getLeftEdge() >= tool.getLeftEdge() && obstacle.getLeftEdge() <= tool.getRightEdge()) || (obstacle.getRightEdge() <= tool.getRightEdge() && obstacle.getRightEdge() >= tool.getLeftEdge());
                if (toolVertexInObstacleEdge || obstacleVertexInToolEdge)
                    isOver.put(obstacle.getTopEdge(), obstacle);
            }
        }
        float max = -100;
        for (float height : isOver.keySet()) {
            if (isOver.get(height).getBottomEdge() < tool.getTopEdge()) {
                max = Math.max(max, height);
            }
        }
        if (max != -100)
            return isOver.get(max);
        return null;
    }

    private float getTerrainEdge(Entity tool, Terrain terrain) {
        for (Tile instance : terrain) {
            if ((tool.getLeftEdge() >= instance.getLeftEdge() && tool.getLeftEdge() <= instance.getRightEdge())
                    || (tool.getRightEdge() <= instance.getRightEdge() && tool.getRightEdge() >= instance.getLeftEdge())
                    || (instance.getLeftEdge() >= tool.getLeftEdge() && instance.getLeftEdge() <= tool.getRightEdge())
                    || (instance.getRightEdge() <= tool.getRightEdge() && instance.getRightEdge() >= tool.getLeftEdge()))
                return terrain.getEdge();
        }
        return -10;
    }

    public PlayerCollisionData getPlayerCollisionData(GameScene scene) {
        EntityCollisionData terrainData = playerCollidesWithTerrain(scene);
        EntityCollisionData ceilingData = playerCollidesWithCeiling(scene);
        ObstacleCollisionData obstacleData = playerCollidesWithObstacle(scene);
        ToolCollisionData toolData = playerCollidesWithTool(scene);
        EntityCollisionData starData = playerCollidesWithStar(scene);
        EntityCollisionData energyData = playerCollidesWithEnergy(scene);
        return new PlayerCollisionData(terrainData, ceilingData, obstacleData, toolData, starData, energyData);
    }

    private EntityCollisionData playerCollidesWithTerrain(GameScene scene) {
        Terrain terrain = scene.getTerrain();
        for (Tile terrainTile : terrain) {
            EntityCollisionData data = detector.playerEntityCollision(scene.getPlayer(), terrainTile);
            if (data.isCollision())
                return data;
        }
        return new EntityCollisionData(null, null);
    }

    private EntityCollisionData playerCollidesWithCeiling(GameScene scene) {
        Ceiling ceiling = scene.getCeiling();
        for (Tile ceilingTile : ceiling) {
            EntityCollisionData data = detector.playerEntityCollision(scene.getPlayer(), ceilingTile);
            if (data.isCollision())
                return data;
        }
        return new EntityCollisionData(null, null);
    }

    private ObstacleCollisionData playerCollidesWithObstacle(GameScene scene) {
        List<EntityCollisionData> collisions = new ArrayList<>();
        synchronized (scene.getObstacles()) {
            for (Obstacle obstacle : scene.getObstacles()) {
                EntityCollisionData data = detector.playerEntityCollision(scene.getPlayer(), obstacle);
                if (data.isCollision())
                    collisions.add(data);
            }
        }
        return new ObstacleCollisionData(collisions);
    }

    private ToolCollisionData playerCollidesWithTool(GameScene scene) {
        List<Tool> tools = new ArrayList<>();
        Circle circle = new Circle(scene.getPlayer());
        synchronized (scene.getTools()) {
            for (Tool tool : scene.getTools()) {
                if (detector.circleToolCollision(circle, tool.getCollisionBounds()))
                    tools.add(tool);
            }
        }
        return new ToolCollisionData(tools);
    }

    private EntityCollisionData playerCollidesWithStar(GameScene scene) {
        synchronized (scene.getStars()) {
            for (Star star : scene.getStars()) {
                EntityCollisionData data = detector.playerEntityCollision(scene.getPlayer(), star);
                if (data.isCollision())
                    return data;
            }
        }
        return new EntityCollisionData(null, null);
    }

    private EntityCollisionData playerCollidesWithEnergy(GameScene scene) {
        if (scene.hasEnergy()) {
            EntityCollisionData data = detector.playerEntityCollision(scene.getPlayer(), scene.getEnergy());
            return data;
        } else
            return new EntityCollisionData(null, null);
    }
}
